@@ -23,6 +23,9 @@ Classes for subcommands of the yum command line interface.
 
 from __future__ import print_function
 from __future__ import unicode_literals
+
+import libdnf
+
 from dnf.cli.option_parser import OptionParser
 from dnf.i18n import _, ucd
 
@@ -33,6 +36,7 @@ import dnf.const
 import dnf.exceptions
 import dnf.i18n
 import dnf.pycomp
+import dnf.transaction
 import dnf.util
 import functools
 import logging
@@ -54,7 +58,7 @@ You can do that by running the command:
 
 
 Alternatively you can specify the url to the key you would like to use
-for a repository in the 'gpgkey' option in a repository section and DNF
+for a repository in the 'gpgkey' option in a repository section and {prog}
 will install it for you.
 
 For more information contact your distribution or package provider.""")
@@ -72,7 +76,7 @@ def _checkGPGKey(base, cli):
     if not base._gpg_key_check():
         for repo in base.repos.iter_enabled():
             if (repo.gpgcheck or repo.repo_gpgcheck) and not repo.gpgkey:
-                logger.critical("\n%s\n", gpg_msg)
+                logger.critical("\n%s\n", gpg_msg.format(prog=dnf.util.MAIN_PROG_UPPER))
                 logger.critical(_("Problem repository: %s"), repo)
                 raise dnf.cli.CliError
 
@@ -93,8 +97,7 @@ def _checkEnabledRepo(base, possible_local_files=()):
         scheme = dnf.pycomp.urlparse.urlparse(lfile)[0]
         if scheme in ('http', 'ftp', 'file', 'https'):
             return
-
-    msg = _('There are no enabled repos.')
+    msg = _('There are no enabled repositories in "{}".').format('", "'.join(base.conf.reposdir))
     raise dnf.cli.CliError(msg)
 
 
@@ -147,6 +150,10 @@ class Command(object):
         """Execute the command."""
         pass
 
+    def run_resolved(self):
+        """Finalize operation after resolvement"""
+        pass
+
     def run_transaction(self):
         """Finalize operations post-transaction."""
         pass
@@ -191,11 +198,11 @@ class InfoCommand(Command):
                              help=_("show only recently changed packages"))
         parser.add_argument('packages', nargs='*', metavar=_('PACKAGE'),
                             choices=cls.pkgnarrows, default=cls.DEFAULT_PKGNARROW,
-                            action=OptionParser.PkgNarrowCallback)
+                            action=OptionParser.PkgNarrowCallback,
+                            help=_("Package name specification"))
 
     def configure(self):
         demands = self.cli.demands
-        demands.fresh_metadata = False
         demands.sack_activation = True
         if self.opts._packages_action:
             self.opts.packages_action = self.opts._packages_action
@@ -210,7 +217,7 @@ class InfoCommand(Command):
             self.opts.packages_action = 'upgrades'
 
     def run(self):
-        self.cli._populate_update_security_filter(self.opts)
+        self.cli._populate_update_security_filter(self.opts, self.base.sack.query())
         return self.base.output_packages('info', self.opts.packages_action,
                                          self.opts.packages)
 
@@ -223,7 +230,7 @@ class ListCommand(InfoCommand):
     summary = _('list a package or groups of packages')
 
     def run(self):
-        self.cli._populate_update_security_filter(self.opts)
+        self.cli._populate_update_security_filter(self.opts, self.base.sack.query())
         return self.base.output_packages('list', self.opts.packages_action,
                                          self.opts.packages)
 
@@ -233,12 +240,13 @@ class ProvidesCommand(Command):
     provides command.
     """
 
-    aliases = ('provides', 'whatprovides')
+    aliases = ('provides', 'whatprovides', 'prov')
     summary = _('find what package provides the given value')
 
     @staticmethod
     def set_argparser(parser):
-        parser.add_argument('dependency', nargs='+', metavar=_('SOME_STRING'))
+        parser.add_argument('dependency', nargs='+', metavar=_('PROVIDE'),
+                            help=_("Provide specification to search for"))
 
     def configure(self):
         demands = self.cli.demands
@@ -247,7 +255,7 @@ class ProvidesCommand(Command):
         demands.sack_activation = True
 
     def run(self):
-        logger.debug("Searching Packages: ")
+        logger.debug(_("Searching Packages: "))
         return self.base.provides(self.opts.dependency)
 
 class CheckUpdateCommand(Command):
@@ -260,18 +268,30 @@ class CheckUpdateCommand(Command):
 
     @staticmethod
     def set_argparser(parser):
+        parser.add_argument('--changelogs', dest='changelogs',
+                            default=False, action='store_true',
+                            help=_('show changelogs before update'))
         parser.add_argument('packages', nargs='*', metavar=_('PACKAGE'))
 
     def configure(self):
         demands = self.cli.demands
         demands.sack_activation = True
         demands.available_repos = True
+        demands.plugin_filtering_enabled = True
+        if self.opts.changelogs:
+            demands.changelogs = True
         _checkEnabledRepo(self.base)
 
     def run(self):
-        self.cli._populate_update_security_filter(self.opts)
+        query = self.base.sack.query().filterm(upgrades_by_priority=True)
+        if self.base.conf.obsoletes:
+            obsoleted = query.union(self.base.sack.query().installed())
+            obsoletes = self.base.sack.query().filter(obsoletes=obsoleted)
+            query = query.union(obsoletes)
+        self.cli._populate_update_security_filter(self.opts, query, cmp_type="gte")
 
-        found = self.base.check_updates(self.opts.packages, print_=True)
+        found = self.base.check_updates(self.opts.packages, print_=True,
+                                        changelogs=self.opts.changelogs)
         if found:
             self.cli.demands.success_exit_status = 100
 
@@ -319,7 +339,7 @@ class RepoPkgsCommand(Command):
 
         def run_on_repo(self):
             """Execute the command with respect to given arguments *cli_args*."""
-            self.cli._populate_update_security_filter(self.opts)
+            self.cli._populate_update_security_filter(self.opts, self.base.sack.query())
             self.base.output_packages('info', self.opts.pkg_specs_action,
                                       self.opts.pkg_specs, self.reponame)
 
@@ -336,6 +356,7 @@ class RepoPkgsCommand(Command):
             demands.root_user = True
 
         def run_on_repo(self):
+            self.cli._populate_update_security_filter(self.opts, self.base.sack.query())
             """Execute the command with respect to given arguments *cli_args*."""
             _checkGPGKey(self.base, self.cli)
 
@@ -354,9 +375,9 @@ class RepoPkgsCommand(Command):
                 for pkg_spec in self.opts.pkg_specs:
                     try:
                         self.base.install(pkg_spec, self.reponame)
-                    except dnf.exceptions.MarkingError:
-                        msg = _('No package %s available.')
-                        logger.info(msg, self.output.term.bold(pkg_spec))
+                    except dnf.exceptions.MarkingError as e:
+                        msg = '{}: {}'.format(e.value, self.base.output.term.bold(pkg_spec))
+                        logger.info(msg)
                     else:
                         done = True
 
@@ -370,7 +391,7 @@ class RepoPkgsCommand(Command):
 
         def run_on_repo(self):
             """Execute the command with respect to given arguments *cli_args*."""
-            self.cli._populate_update_security_filter(self.opts)
+            self.cli._populate_update_security_filter(self.opts, self.base.sack.query())
             self.base.output_packages('list', self.opts.pkg_specs_action,
                                       self.opts.pkg_specs, self.reponame)
 
@@ -415,9 +436,9 @@ class RepoPkgsCommand(Command):
                     except dnf.exceptions.PackagesNotAvailableError as err:
                         for pkg in err.packages:
                             xmsg = ''
-                            yumdb_info = self.base._yumdb.get_package(pkg)
-                            if 'from_repo' in yumdb_info:
-                                xmsg = _(' (from %s)') % yumdb_info.from_repo
+                            pkgrepo = self.base.history.repo(pkg)
+                            if pkgrepo:
+                                xmsg = _(' (from %s)') % pkgrepo
                             msg = _('Installed package %s%s not available.')
                             logger.info(msg, self.output.term.bold(pkg), xmsg)
                     except dnf.exceptions.MarkingError:
@@ -472,9 +493,9 @@ class RepoPkgsCommand(Command):
                     except dnf.exceptions.PackagesNotAvailableError as err:
                         for pkg in err.packages:
                             xmsg = ''
-                            yumdb_info = self.base._yumdb.get_package(pkg)
-                            if 'from_repo' in yumdb_info:
-                                xmsg = _(' (from %s)') % yumdb_info.from_repo
+                            pkgrepo = self.base.history.repo(pkg)
+                            if pkgrepo:
+                                xmsg = _(' (from %s)') % pkgrepo
                             msg = _('Installed package %s%s not available.')
                             logger.info(msg, self.output.term.bold(pkg), xmsg)
                     except dnf.exceptions.MarkingError:
@@ -535,10 +556,10 @@ class RepoPkgsCommand(Command):
 
             subject = dnf.subject.Subject(pkg_spec)
             matches = subject.get_best_query(self.cli.base.sack)
-            yumdb = self.cli.base._yumdb
+            history = self.cli.base.history
             installed = [
                 pkg for pkg in matches.installed()
-                if yumdb.get_package(pkg).get('from_repo') == reponame]
+                if history.repo(pkg) == reponame]
             if not installed:
                 raise dnf.exceptions.PackagesNotInstalledError(
                     'no package matched', pkg_spec)
@@ -660,13 +681,13 @@ class RepoPkgsCommand(Command):
                 for pkg_spec in self.opts.pkg_specs:
                     try:
                         self.base.remove(pkg_spec, self.reponame)
-                    except dnf.exceptions.MarkingError:
-                        logger.info(_('No match for argument: %s'), pkg_spec)
+                    except dnf.exceptions.MarkingError as e:
+                        logger.info(str(e))
                     else:
                         done = True
 
             if not done:
-                raise dnf.exceptions.Error(_('No packages marked for removal.'))
+                logger.warning(_('No packages marked for removal.'))
 
     class UpgradeSubCommand(Command):
         """Implementation of the upgrade sub-command."""
@@ -722,16 +743,6 @@ class RepoPkgsCommand(Command):
             alias: subcmd for subcmd in subcmd_objs for alias in subcmd.aliases}
 
     def set_argparser(self, parser):
-        subcommand_choices = [subcmd.aliases[0] for subcmd in self.SUBCMDS]
-        super(OptionParser, parser).add_argument(
-            'reponame', nargs=1, action=OptionParser._RepoCallbackEnable,
-            metavar=_('REPO'))
-        parser.add_argument('subcmd', nargs=1, choices=subcommand_choices)
-        DEFAULT_PKGNARROW = 'all'
-        pkgnarrows = {DEFAULT_PKGNARROW, 'installed', 'available',
-                      'autoremove', 'extras', 'obsoletes', 'recent',
-                      'upgrades'}
-
         narrows = parser.add_mutually_exclusive_group()
         narrows.add_argument('--all', dest='_pkg_specs_action',
                              action='store_const', const='all', default=None,
@@ -757,9 +768,21 @@ class RepoPkgsCommand(Command):
         narrows.add_argument('--recent', dest='_pkg_specs_action',
                              action='store_const', const='recent',
                              help=_("show only recently changed packages"))
+
+        parser.add_argument(
+            'reponame', nargs=1, action=OptionParser._RepoCallbackEnable,
+            metavar=_('REPOID'), help=_("Repository ID"))
+        subcommand_choices = [subcmd.aliases[0] for subcmd in self.SUBCMDS]
+        parser.add_argument('subcmd', nargs=1, metavar="SUBCOMMAND",
+                            choices=subcommand_choices, help=", ".join(subcommand_choices))
+        DEFAULT_PKGNARROW = 'all'
+        pkgnarrows = {DEFAULT_PKGNARROW, 'installed', 'available',
+                      'autoremove', 'extras', 'obsoletes', 'recent',
+                      'upgrades'}
         parser.add_argument('pkg_specs', nargs='*', metavar=_('PACKAGE'),
                             choices=pkgnarrows, default=DEFAULT_PKGNARROW,
-                            action=OptionParser.PkgNarrowCallback)
+                            action=OptionParser.PkgNarrowCallback,
+                            help=_("Package specification"))
 
     def configure(self):
         """Verify whether the command can run with given arguments."""
@@ -787,7 +810,9 @@ class HelpCommand(Command):
 
     @staticmethod
     def set_argparser(parser):
-        parser.add_argument('cmd', nargs='?', metavar=_('COMMAND'))
+        parser.add_argument('cmd', nargs='?', metavar=_('COMMAND'),
+                            help=_("{prog} command to get help for").format(
+                                prog=dnf.util.MAIN_PROG_UPPER))
 
     def run(self):
         if (not self.opts.cmd
@@ -802,43 +827,58 @@ class HistoryCommand(Command):
     history command.
     """
 
-    aliases = ('history',)
+    aliases = ('history', 'hist')
     summary = _('display, or use, the transaction history')
 
+    _CMDS = ['list', 'info', 'redo', 'undo', 'rollback', 'userinstalled']
+
     transaction_ids = set()
     merged_transaction_ids = set()
 
     @staticmethod
     def set_argparser(parser):
-        cmds = ['list', 'info', 'redo', 'undo', 'rollback', 'userinstalled']
-        parser.add_argument('transactions', nargs='*',
-                            choices=cmds, default=cmds[0],
-                            action=OptionParser.PkgNarrowCallback,
-                            metavar="[%s]" % "|".join(cmds))
+        parser.add_argument('transactions_action', nargs='?', metavar="COMMAND",
+                            help="Available commands: {} (default), {}".format(
+                                HistoryCommand._CMDS[0],
+                                ", ".join(HistoryCommand._CMDS[1:])))
+        parser.add_argument('transactions', nargs='*', metavar="TRANSACTION",
+                            help="Transaction ID (<number>, 'last' or 'last-<number>' "
+                                 "for one transaction, <transaction-id>..<transaction-id> "
+                                 "for range)")
 
     def configure(self):
+        if not self.opts.transactions_action:
+            # no positional argument given
+            self.opts.transactions_action = self._CMDS[0]
+        elif self.opts.transactions_action not in self._CMDS:
+            # first positional argument is not a command
+            self.opts.transactions.insert(0, self.opts.transactions_action)
+            self.opts.transactions_action = self._CMDS[0]
+
         require_one_transaction_id = False
         require_one_transaction_id_msg = _("Found more than one transaction ID.\n"
                                            "'{}' requires one transaction ID or package name."
                                            ).format(self.opts.transactions_action)
         demands = self.cli.demands
         if self.opts.transactions_action in ['redo', 'undo', 'rollback']:
+            demands.root_user = True
             require_one_transaction_id = True
             if not self.opts.transactions:
-                logger.critical(_('No transaction ID or package name given.'))
-                raise dnf.cli.CliError
+                msg = _('No transaction ID or package name given.')
+                logger.critical(msg)
+                raise dnf.cli.CliError(msg)
             elif len(self.opts.transactions) > 1:
                 logger.critical(require_one_transaction_id_msg)
-                raise dnf.cli.CliError
+                raise dnf.cli.CliError(require_one_transaction_id_msg)
             demands.available_repos = True
             _checkGPGKey(self.base, self.cli)
         else:
             demands.fresh_metadata = False
         demands.sack_activation = True
-        demands.root_user = True
-        if not os.access(self.base.history._db_file, os.R_OK):
-            logger.critical(_("You don't have access to the history DB."))
-            raise dnf.cli.CliError
+        if not os.access(self.base.history.path, os.R_OK):
+            msg = _("You don't have access to the history DB.")
+            logger.critical(msg)
+            raise dnf.cli.CliError(msg)
         self.transaction_ids = self._args2transaction_ids(self.merged_transaction_ids,
                                                           require_one_transaction_id,
                                                           require_one_transaction_id_msg)
@@ -866,22 +906,24 @@ class HistoryCommand(Command):
         print('Repeating transaction %u, from %s' % (old.tid, tm))
         self.output.historyInfoCmdPkgsAltered(old)
 
-        converter = dnf.history.TransactionConverter(self.base.sack)
-        history = dnf.history.open_history(self.base.history)
-        operations = history.transaction_nevra_ops(old.tid)
+        for i in old.packages():
+            pkgs = list(self.base.sack.query().filter(nevra=str(i), reponame=i.from_repo))
+            if i.action in dnf.transaction.FORWARD_ACTIONS:
+                if not pkgs:
+                    logger.info(_('No package %s available.'),
+                    self.output.term.bold(ucd(str(i))))
+                    return 1, ['An operation cannot be redone']
+                pkg = pkgs[0]
+                self.base.install(str(pkg))
+            elif i.action == libdnf.transaction.TransactionItemAction_REMOVE:
+                if not pkgs:
+                    # package was removed already, we can skip removing it again
+                    continue
+                pkg = pkgs[0]
+                self.base.remove(str(pkg))
 
-        try:
-            self.base.transaction = converter.convert(operations, 'history')
-        except dnf.exceptions.PackagesNotInstalledError as err:
-            logger.info(_('No package %s installed.'),
-                        self.output.term.bold(ucd(err.pkg_spec)))
-            return 1, ['An operation cannot be redone']
-        except dnf.exceptions.PackagesNotAvailableError as err:
-            logger.info(_('No package %s available.'),
-                        self.output.term.bold(ucd(err.pkg_spec)))
-            return 1, ['An operation cannot be redone']
-        else:
-            return 2, ['Repeating transaction %u' % (old.tid,)]
+        self.base.resolve()
+        self.base.do_transaction()
 
     def _hcmd_undo(self, extcmds):
         try:
@@ -925,8 +967,8 @@ class HistoryCommand(Command):
                           "Use '<transaction-id>..<transaction-id>'."
                           ).format(t))
                     raise dnf.cli.CliError
-                cant_convert_msg = "Can't convert '{}' to transaction ID.\n" \
-                                   "Use '<integer>', 'last', 'last-<positive-integer>'."
+                cant_convert_msg = _("Can't convert '{}' to transaction ID.\n"
+                                     "Use '<number>', 'last', 'last-<number>'.")
                 try:
                     begin_transaction_id = str2transaction_id(begin_transaction_id)
                 except ValueError:
@@ -967,9 +1009,10 @@ class HistoryCommand(Command):
     def run(self):
         vcmd = self.opts.transactions_action
 
-        if vcmd == 'list':
+        ret = None
+        if vcmd == 'list' and (self.transaction_ids or not self.opts.transactions):
             ret = self.output.historyListCmd(self.transaction_ids)
-        elif vcmd == 'info':
+        elif vcmd == 'info' and (self.transaction_ids or not self.opts.transactions):
             ret = self.output.historyInfoCmd(self.transaction_ids, self.opts.transactions,
                                              self.merged_transaction_ids)
         elif vcmd == 'undo':
