@@ -31,13 +31,22 @@ STATE_ENABLED = libdnf.module.ModulePackageContainer.ModuleState_ENABLED
 STATE_DISABLED = libdnf.module.ModulePackageContainer.ModuleState_DISABLED
 STATE_UNKNOWN = libdnf.module.ModulePackageContainer.ModuleState_UNKNOWN
 MODULE_TABLE_HINT = _("\n\nHint: [d]efault, [e]nabled, [x]disabled, [i]nstalled")
+MODULE_INFO_TABLE_HINT = _("\n\nHint: [d]efault, [e]nabled, [x]disabled, [i]nstalled, [a]ctive")
+
+
+def _profile_comparison_key(profile):
+    return profile.getName()
+
 
 class ModuleBase(object):
+    # :api
 
     def __init__(self, base):
+        # :api
         self.base = base
 
     def enable(self, module_specs):
+        # :api
         no_match_specs, error_specs, solver_errors, module_dicts = \
             self._resolve_specs_enable_update_sack(module_specs)
         for spec, (nsvcap, module_dict) in module_dicts.items():
@@ -47,49 +56,79 @@ class ModuleBase(object):
         if no_match_specs or error_specs or solver_errors:
             raise dnf.exceptions.MarkingErrors(no_match_group_specs=no_match_specs,
                                                error_group_specs=error_specs,
-                                               module_debsolv_errors=solver_errors)
+                                               module_depsolv_errors=solver_errors)
 
     def disable(self, module_specs):
+        # :api
         no_match_specs, solver_errors = self._modules_reset_or_disable(module_specs, STATE_DISABLED)
         if no_match_specs or solver_errors:
             raise dnf.exceptions.MarkingErrors(no_match_group_specs=no_match_specs,
-                                               module_debsolv_errors=solver_errors)
+                                               module_depsolv_errors=solver_errors)
 
     def install(self, module_specs, strict=True):
+        # :api
         no_match_specs, error_specs, solver_errors, module_dicts = \
             self._resolve_specs_enable_update_sack(module_specs)
 
         # <package_name, set_of_spec>
+        fail_safe_repo = hawkey.MODULE_FAIL_SAFE_REPO_NAME
         install_dict = {}
         install_set_artefacts = set()
+        fail_safe_repo_used = False
         for spec, (nsvcap, moduledict) in module_dicts.items():
             for name, streamdict in moduledict.items():
                 for stream, module_list in streamdict.items():
                     install_module_list = [x for x in module_list
                                            if self.base._moduleContainer.isModuleActive(x.getId())]
                     if not install_module_list:
+                        logger.error(_("All matches for argument '{0}' in module '{1}:{2}' are not "
+                                       "active").format(spec, name, stream))
                         error_specs.append(spec)
                         continue
                     profiles = []
                     latest_module = self._get_latest(install_module_list)
+                    if latest_module.getRepoID() == fail_safe_repo:
+                        msg = _(
+                            "Installing module '{0}' from Fail-Safe repository {1} is not allowed")
+                        logger.critical(msg.format(latest_module.getNameStream(), fail_safe_repo))
+                        fail_safe_repo_used = True
                     if nsvcap.profile:
                         profiles.extend(latest_module.getProfiles(nsvcap.profile))
                         if not profiles:
-                            logger.error(_("Unable to match profile in argument {}").format(spec))
-                            error_specs.append(spec)
+                            available_profiles = latest_module.getProfiles()
+                            if available_profiles:
+                                profile_names = ", ".join(sorted(
+                                    [profile.getName() for profile in available_profiles]))
+                                msg = _("Unable to match profile for argument {}. Available "
+                                        "profiles for '{}:{}': {}").format(
+                                    spec, name, stream, profile_names)
+                            else:
+                                msg = _("Unable to match profile for argument {}").format(spec)
+                            logger.error(msg)
+                            no_match_specs.append(spec)
                             continue
                     else:
                         profiles_strings = self.base._moduleContainer.getDefaultProfiles(
                             name, stream)
                         if not profiles_strings:
-                            logger.error(_("No default profiles for module {}:{}").format(
-                                name, stream))
+                            available_profiles = latest_module.getProfiles()
+                            if available_profiles:
+                                profile_names = ", ".join(sorted(
+                                    [profile.getName() for profile in available_profiles]))
+                                msg = _("No default profiles for module {}:{}. Available profiles"
+                                        ": {}").format(
+                                    name, stream, profile_names)
+                            else:
+                                msg = _("No profiles for module {}:{}").format(name, stream)
+                            logger.error(msg)
+                            error_specs.append(spec)
                         for profile in set(profiles_strings):
                             module_profiles = latest_module.getProfiles(profile)
                             if not module_profiles:
                                 logger.error(
-                                    _("Profile {} not matched for module {}:{}").format(
+                                    _("Default profile {} not available in module {}:{}").format(
                                         profile, name, stream))
+                                error_specs.append(spec)
 
                             profiles.extend(module_profiles)
                     for profile in profiles:
@@ -98,9 +137,18 @@ class ModuleBase(object):
                             install_dict.setdefault(pkg_name, set()).add(spec)
                     for module in install_module_list:
                         install_set_artefacts.update(module.getArtifacts())
+        if fail_safe_repo_used:
+            raise dnf.exceptions.Error(_(
+                "Installing module from Fail-Safe repository is not allowed"))
         install_base_query = self.base.sack.query().filterm(
             nevra_strict=install_set_artefacts).apply()
 
+        # add hot-fix packages
+        hot_fix_repos = [i.id for i in self.base.repos.iter_enabled() if i.module_hotfixes]
+        hotfix_packages = self.base.sack.query().filterm(reponame=hot_fix_repos).filterm(
+            name=install_dict.keys())
+        install_base_query = install_base_query.union(hotfix_packages)
+
         for pkg_name, set_specs in install_dict.items():
             query = install_base_query.filter(name=pkg_name)
             if not query:
@@ -119,16 +167,20 @@ class ModuleBase(object):
         if no_match_specs or error_specs or solver_errors:
             raise dnf.exceptions.MarkingErrors(no_match_group_specs=no_match_specs,
                                                error_group_specs=error_specs,
-                                               module_debsolv_errors=solver_errors)
+                                               module_depsolv_errors=solver_errors)
 
     def reset(self, module_specs):
+        # :api
         no_match_specs, solver_errors = self._modules_reset_or_disable(module_specs, STATE_UNKNOWN)
         if no_match_specs:
             raise dnf.exceptions.MarkingErrors(no_match_group_specs=no_match_specs,
-                                               module_debsolv_errors=solver_errors)
+                                               module_depsolv_errors=solver_errors)
 
     def upgrade(self, module_specs):
+        # :api
         no_match_specs = []
+        fail_safe_repo = hawkey.MODULE_FAIL_SAFE_REPO_NAME
+        fail_safe_repo_used = False
 
         for spec in module_specs:
             module_list, nsvcap = self._get_modules(spec)
@@ -147,31 +199,40 @@ class ModuleBase(object):
                     upgrade_package_set.update(self._get_package_name_set_and_remove_profiles(
                         module_list_from_dict, nsvcap))
                     latest_module = self._get_latest(module_list_from_dict)
-                    installed_profiles_strings = set(
-                        self.base._moduleContainer.getInstalledProfiles(latest_module.getName()))
-                    if not installed_profiles_strings:
-                        continue
+                    if latest_module.getRepoID() == fail_safe_repo:
+                        msg = _(
+                            "Upgrading module '{0}' from Fail-Safe repository {1} is not allowed")
+                        logger.critical(msg.format(latest_module.getNameStream(), fail_safe_repo))
+                        fail_safe_repo_used = True
                     if nsvcap.profile:
                         profiles_set = latest_module.getProfiles(nsvcap.profile)
                         if not profiles_set:
                             continue
                         for profile in profiles_set:
-                            if profile.getName() in installed_profiles_strings:
-                                upgrade_package_set.update(profile.getContent())
+                            upgrade_package_set.update(profile.getContent())
                     else:
-                        for profile_string in installed_profiles_strings:
-                            for profile in latest_module.getProfiles(profile_string):
-                                upgrade_package_set.update(profile.getContent())
+                        for profile in latest_module.getProfiles():
+                            upgrade_package_set.update(profile.getContent())
+                        for artefact in latest_module.getArtifacts():
+                            subj = hawkey.Subject(artefact)
+                            for nevra_obj in subj.get_nevra_possibilities(
+                                    forms=[hawkey.FORM_NEVRA]):
+                                upgrade_package_set.add(nevra_obj.name)
+
             if not upgrade_package_set:
                 logger.error(_("Unable to match profile in argument {}").format(spec))
-            query = self.base.sack.query().available().filterm(name=upgrade_package_set)
+            query = self.base.sack.query().filterm(name=upgrade_package_set)
             if query:
                 sltr = dnf.selector.Selector(self.base.sack)
                 sltr.set(pkg=query)
                 self.base._goal.upgrade(select=sltr)
+        if fail_safe_repo_used:
+            raise dnf.exceptions.Error(_(
+                "Upgrading module from Fail-Safe repository is not allowed"))
         return no_match_specs
 
     def remove(self, module_specs):
+        # :api
         no_match_specs = []
         remove_package_set = set()
 
@@ -199,7 +260,12 @@ class ModuleBase(object):
                     self.base._remove_if_unneeded(query)
         return no_match_specs
 
+    def get_modules(self, module_spec):
+        # :api
+        return self._get_modules(module_spec)
+
     def _get_modules(self, module_spec):
+        # used by ansible (lib/ansible/modules/packaging/os/dnf.py)
         subj = hawkey.Subject(module_spec)
         for nsvcap in subj.nsvcap_possibilities():
             name = nsvcap.name if nsvcap.name else ""
@@ -219,7 +285,7 @@ class ModuleBase(object):
         if module_list:
             latest = module_list[0]
             for module in module_list[1:]:
-                if module.getVersion() > latest.getVersion():
+                if module.getVersionNum() > latest.getVersionNum():
                     latest = module
         return latest
 
@@ -270,15 +336,23 @@ class ModuleBase(object):
                 logger.error(ucd(e))
                 logger.error(_("Unable to resolve argument {}").format(spec))
         hot_fix_repos = [i.id for i in self.base.repos.iter_enabled() if i.module_hotfixes]
-        solver_errors = self.base.sack.filter_modules(
-            self.base._moduleContainer, hot_fix_repos, self.base.conf.installroot, None,
-            self.base.conf.debug_solver
-        )
-        for nsvcap, moduleDict in module_dicts.values():
+        try:
+            solver_errors = self.base.sack.filter_modules(
+                self.base._moduleContainer, hot_fix_repos, self.base.conf.installroot,
+                self.base.conf.module_platform_id, update_only=True,
+                debugsolver=self.base.conf.debug_solver)
+        except hawkey.Exception as e:
+            raise dnf.exceptions.Error(ucd(e))
+        for spec, (nsvcap, moduleDict) in module_dicts.items():
             for streamDict in moduleDict.values():
                 for modules in streamDict.values():
-                    self.base._moduleContainer.enableDependencyTree(
-                        libdnf.module.VectorModulePackagePtr(modules))
+                    try:
+                        self.base._moduleContainer.enableDependencyTree(
+                            libdnf.module.VectorModulePackagePtr(modules))
+                    except RuntimeError as e:
+                        error_spec.append(spec)
+                        logger.error(ucd(e))
+                        logger.error(_("Unable to resolve argument {}").format(spec))
         return no_match_specs, error_spec, solver_errors, module_dicts
 
     def _modules_reset_or_disable(self, module_specs, to_state):
@@ -289,9 +363,9 @@ class ModuleBase(object):
                 logger.error(_("Unable to resolve argument {}").format(spec))
                 no_match_specs.append(spec)
                 continue
-            if nsvcap.profile:
-                logger.info(_("Ignoring unnecessary profile: '{}/{}'").format(
-                    nsvcap.name, nsvcap.profile))
+            if nsvcap.stream or nsvcap.version or nsvcap.context or nsvcap.arch or nsvcap.profile:
+                logger.info(_("Only module name is required. "
+                              "Ignoring unneeded information in argument: '{}'").format(spec))
             module_names = set()
             for module in module_list:
                 module_names.add(module.getName())
@@ -302,10 +376,13 @@ class ModuleBase(object):
                     self.base._moduleContainer.disable(name)
 
         hot_fix_repos = [i.id for i in self.base.repos.iter_enabled() if i.module_hotfixes]
-        solver_errors = self.base.sack.filter_modules(
-            self.base._moduleContainer, hot_fix_repos, self.base.conf.installroot,
-            self.base.conf.module_platform_id, update_only=True,
-            debugsolver=self.base.conf.debug_solver)
+        try:
+            solver_errors = self.base.sack.filter_modules(
+                self.base._moduleContainer, hot_fix_repos, self.base.conf.installroot,
+                self.base.conf.module_platform_id, update_only=True,
+                debugsolver=self.base.conf.debug_solver)
+        except hawkey.Exception as e:
+            raise dnf.exceptions.Error(ucd(e))
         return no_match_specs, solver_errors
 
     def _get_package_name_set_and_remove_profiles(self, module_list, nsvcap, remove=False):
@@ -348,26 +425,26 @@ class ModuleBase(object):
                 lines = OrderedDict()
                 lines["Name"] = module.getFullIdentifier()
 
-                for profile in module.getProfiles():
+                for profile in sorted(module.getProfiles(), key=_profile_comparison_key):
                     lines[profile.getName()] = "\n".join(
                         [pkgName for pkgName in profile.getContent()])
 
                 output.add(self._create_simple_table(lines).toString())
         return "\n\n".join(sorted(output))
 
-    def _profile_report_formater(self, modulePackage, default_profiles, enabled_str):
+    def _profile_report_formatter(self, modulePackage, default_profiles, enabled_str):
         installed_profiles = self.base._moduleContainer.getInstalledProfiles(
             modulePackage.getName())
         available_profiles = modulePackage.getProfiles()
         profiles_str = ""
-        for profile in available_profiles:
+        for profile in sorted(available_profiles, key=_profile_comparison_key):
             profiles_str += "{}{}".format(
                 profile.getName(), " [d]" if profile.getName() in default_profiles else "")
             profiles_str += " [i], " if profile.getName() in installed_profiles and enabled_str \
                 else ", "
         return profiles_str[:-2]
 
-    def _module_strs_formater(self, modulePackage):
+    def _module_strs_formatter(self, modulePackage, markActive=False):
         default_str = ""
         enabled_str = ""
         disabled_str = ""
@@ -382,6 +459,10 @@ class ModuleBase(object):
             if not default_str:
                 disabled_str = " "
             disabled_str += "[x]"
+        if markActive and self.base._moduleContainer.isModuleActive(modulePackage):
+            if not default_str:
+                disabled_str = " "
+            disabled_str += "[a]"
         return default_str, enabled_str, disabled_str
 
     def _get_info(self, module_specs):
@@ -396,11 +477,12 @@ class ModuleBase(object):
                 logger.info(_("Ignoring unnecessary profile: '{}/{}'").format(
                     nsvcap.name, nsvcap.profile))
             for modulePackage in module_list:
-                default_str, enabled_str, disabled_str = self._module_strs_formater(modulePackage)
+                default_str, enabled_str, disabled_str = self._module_strs_formatter(
+                    modulePackage, markActive=True)
                 default_profiles = self.base._moduleContainer.getDefaultProfiles(
                     modulePackage.getName(), modulePackage.getStream())
 
-                profiles_str = self._profile_report_formater(
+                profiles_str = self._profile_report_formatter(
                     modulePackage, default_profiles, enabled_str)
 
                 lines = OrderedDict()
@@ -409,16 +491,23 @@ class ModuleBase(object):
                                   disabled_str
                 lines["Version"] = modulePackage.getVersion()
                 lines["Context"] = modulePackage.getContext()
+                lines["Architecture"] = modulePackage.getArch()
                 lines["Profiles"] = profiles_str
                 lines["Default profiles"] = " ".join(default_profiles)
                 lines["Repo"] = modulePackage.getRepoID()
                 lines["Summary"] = modulePackage.getSummary()
                 lines["Description"] = modulePackage.getDescription()
+                req_set = set()
+                for req in modulePackage.getModuleDependencies():
+                    for require_dict in req.getRequires():
+                        for mod_require, stream in require_dict.items():
+                            req_set.add("{}:[{}]".format(mod_require, ",".join(stream)))
+                lines["Requires"] = "\n".join(sorted(req_set))
                 lines["Artifacts"] = "\n".join(sorted(modulePackage.getArtifacts()))
                 output.add(self._create_simple_table(lines).toString())
         str_table = "\n\n".join(sorted(output))
         if str_table:
-            str_table += MODULE_TABLE_HINT
+            str_table += MODULE_INFO_TABLE_HINT
         return str_table
 
     @staticmethod
@@ -434,10 +523,11 @@ class ModuleBase(object):
         column_value.setNewlineWrapFunction()
 
         for line_name, value in lines.items():
-            if value:
-                line = table.newLine()
-                line.getColumnCell(column_name).setData(line_name)
-                line.getColumnCell(column_value).setData(str(value))
+            if value is None:
+                value = ""
+            line = table.newLine()
+            line.getColumnCell(column_name).setData(line_name)
+            line.getColumnCell(column_value).setData(str(value))
 
         return table
 
@@ -463,10 +553,14 @@ class ModuleBase(object):
         output = set()
         modulePackages = self.base._moduleContainer.getModulePackages()
         baseQuery = self.base.sack.query().filterm(empty=True).apply()
+        getBestInitQuery = self.base.sack.query(flags=hawkey.IGNORE_MODULAR_EXCLUDES)
+
         for spec in rpm_specs:
             subj = dnf.subject.Subject(spec)
             baseQuery = baseQuery.union(subj.get_best_query(
-                self.base.sack, with_nevra=True, with_provides=False, with_filenames=False))
+                self.base.sack, with_nevra=True, with_provides=False, with_filenames=False,
+                query=getBestInitQuery))
+
         baseQuery.apply()
 
         for modulePackage in modulePackages:
@@ -478,12 +572,12 @@ class ModuleBase(object):
                 for pkg in query:
                     string_output = ""
                     profiles = []
-                    for profile in modulePackage.getProfiles():
+                    for profile in sorted(modulePackage.getProfiles(), key=_profile_comparison_key):
                         if pkg.name in profile.getContent():
                             profiles.append(profile.getName())
                     lines = OrderedDict()
                     lines["Module"] = modulePackage.getFullIdentifier()
-                    lines["Profiles"] = " ".join(profiles)
+                    lines["Profiles"] = " ".join(sorted(profiles))
                     lines["Repo"] = modulePackage.getRepoID()
                     lines["Summary"] = modulePackage.getSummary()
 
@@ -521,10 +615,11 @@ class ModuleBase(object):
                     else:
                         modulePackage = nameStreamArch[0]
                 line = table.newLine()
-                default_str, enabled_str, disabled_str = self._module_strs_formater(modulePackage)
+                default_str, enabled_str, disabled_str = self._module_strs_formatter(
+                    modulePackage, markActive=False)
                 default_profiles = self.base._moduleContainer.getDefaultProfiles(
                     modulePackage.getName(), modulePackage.getStream())
-                profiles_str = self._profile_report_formater(modulePackage, default_profiles,
+                profiles_str = self._profile_report_formatter(modulePackage, default_profiles,
                                                              enabled_str)
                 line.getColumnCell(column_name).setData(modulePackage.getName())
                 line.getColumnCell(
@@ -550,7 +645,10 @@ class ModuleBase(object):
         table = self._create_and_fill_table(latest)
         current_repo_id_index = 0
         already_printed_lines = 0
-        repo_name = self.base.repos[latest[0][0][0].getRepoID()].name
+        try:
+            repo_name = self.base.repos[latest[0][0][0].getRepoID()].name
+        except KeyError:
+            repo_name = latest[0][0][0].getRepoID()
         versions = len(latest[0])
         header = self._format_header(table)
         str_table = self._format_repoid(repo_name)
@@ -559,7 +657,12 @@ class ModuleBase(object):
             if versions + already_printed_lines <= i:
                 already_printed_lines += versions
                 current_repo_id_index += 1
-                repo_name = self.base.repos[latest[current_repo_id_index][0][0].getRepoID()].name
+                # Fail-Safe repository is not in self.base.repos
+                try:
+                    repo_name = self.base.repos[
+                        latest[current_repo_id_index][0][0].getRepoID()].name
+                except KeyError:
+                    repo_name = latest[current_repo_id_index][0][0].getRepoID()
                 versions = len(latest[current_repo_id_index])
                 str_table += "\n"
                 str_table += self._format_repoid(repo_name)
